TON blockchain contest
======================

Greetings. My name is Nickolay Platonov, I'm participating in the TON blockchain contest. I only had a few free days for it, so my application does not strictly follows the contest conditions. Therefore, I'm not pretending on any kind of prize, but still hope my application is relevant and of interest for the TON team.

It consist from the two reference implementations of the multi signature wallet. The 1st one is written in the [TypeScript] and 2nd in [Idris] - a functional language, which can be seen as Haskell with dependent types. 

I've used the TypeScript implementation for the quick prototyping and understanding the data domain. In the Idris implementation I'm exploring the possibilities of the static (compile-time) guarantees for the smart contract code.


Multi-signature wallet (TypeScript)
-----------------------------------

This implementation is covered with the test suite. To run the tests (assuming you have Node.js installed):
- complete `npm install` in the `muti_sig_wallet_typescript` folder
- run `npx tsc`
- run `npx siesta ./tests`
- or, open the `http://YOUR_LOCAL_WEBSERVER_PATH/muti_sig_wallet_typescript/test/index.html` in the browser and double click the `010_wallet.t.ts` file in the left tree.

I will quickly go through the state flow in this implementation. The TypeScript syntax should be common enough to serve as a "lingua franca" for the developers with the C++ background.

It is assumed, that an instance of the `MultiSigWallet` is already deployed somewhere in the TON VM. This class has the following state:
- `owners` - a set of wallet owners
- `quorum` - a number of owners signatures, required to release a transfer order
- `stateClock` - a logical clock for the state of the wallet
- `transferOrders` - a map of not yet released transfer orders (`TransferOrder`).

Few trivial invariants are enforced during wallet creation - at least one owner should be specified, quorum should be specified, quorum should not exceed the number of owners, etc.

The flow starts when some user creates a transfer order request (instance of `TransferOrderRequest`). I think, a common intuition is, that transfer order maintains own identity besides its data, therefore, I've included a GUID into the `TransferOrderRequest`, to be able to distinguish between several "twin" orders with the same amount and destination.

A transfer initiator should sign the `TransferOrderRequest` with its public key (`newTransferOrder` helper method), which produces a `TransferOrder`. The serialized instance of the latter can be distributed offline between the wallet owners, accumulating the signatures in the `signedBy` property.

At any point of time, `TransferOrder` (which possibly has not reached the quorum yet), can be sent to the wallet, wrapped in a `Message`. `Message` is parameterized by the payload type, having a single mandatory field `stateClock`, which guards the wallet state and guarantees consistency (as suggested by the "simple-wallet" implementation). 

    Should be noted, that such consistency maintenance scheme is can hardly be seen as optimal, as the consequent flow of the "state clock" can be easily interrupted by another actor, thus it should only be used for trivial applications. Probably, a more advanced scheme, like the STM (software transactional memory) can be employed instead.

The message sent will be processed by the `addOrMergeTransfer` method of the wallet. The interesting part of it is the verification of the incoming signatures. We want to "guard" the `signedBy` property of the `TransferOrder` with the proof, that signature belongs to one of the wallet's owners. For this purpose we create a private method `addSignature`, which "builds a proof" and then extends the `signedBy` property. Of course, since everything happens in the run-time, "building a proof" reduces to a mere boolean check. 

If the transfer order has accumulated the quorum, it is "released", otherwise it remains in the `this.transferOrders` map (the latter is garbage collected at every incoming request). At this point a transfer is considered completed and removed from the internal wallet's state.

In addition to this data flow, a few self-explanatory information retrieval helpers are added: `getOrders`, `getOrdersSignedBy`, `getOrdersNotSignedBy`. One can note, that such helpers do not advance the state clock, since no mutations happens.

## Conclusion
In general, the implementation is trivial. I simply had not enough time to translate it to the suggested TON language - FunC (which just would be a tedious trial and error procedure). To be honest, I doubt that FunC will gain developers mindshare, as it is too low-level and forces developers to think in terms of assembler language. Other, higher-level languages (like Solidity) will probably win the mindshare. It is also strange that `http://ton.dev` website claims to have the Solidity to TVM compiler - if so, why the contest terms were suggesting to use FunC? Well, thats probably something internal to the TON project, and I'm just proceeding to the next implementation.


Multi-signature wallet (Idris)
------------------------------

This reference implementation explores the possibility of enforcing static code safety. Since the cost of the sofware bug can be unacceptably high for the financial domain, users wants to have maximum assurance, that smart contracts, they are dealing with, operates strictly according to the specification.

It is of course possible to encode the specification as the run-time checks in the imperative language, however, software development practice shows, that such checks never provide necessary level of assurance. Therefore, there's an interest of expressing and enforcing the specification statically, at the program compilation time. Functional paradigm offers much more facilities for this, especially the languages with the dependent types, like [Idris].

Here we create a dependent type, that holds the state of the wallet:

    data MultiSigWalletState : (address : Address) -> (balance : NanoGram) -> (owners : Vect (S n) PublicIdentity) -> (quorum : NonZeroNat) -> Type where
        MkMultiSigWalletState : MultiSigWalletState address balance owners quorum

Then we define a type-level predicate, describing the state when quorum about the withdrawal operation has been reached:

    data QuorumReached :
        (wallet : MultiSigWalletState address balance owners quorum)
        -> (signatures : Vect (S k) PublicIdentity)
        -> Type
    where
        IfAndOnlyIf :
            (wallet : MultiSigWalletState address balance owners quorum)
            -> (allSignaturesBelongsToOwners : ForEveryElementVect signatures (SignatureBelongsToOwners owners))
            -> (enoughSignatures : GTE (S k) (extractNonZeroNat quorum))
            -> QuorumReached wallet signatures
            
We also define a decision procedure for it: 

    quorumReached :
        (wallet : MultiSigWalletState address balance owners quorum)
        -> (signatures : Vect (S k) PublicIdentity)
        -> {auto enoughSignatures : GTE (S k) (extractNonZeroNat quorum) }
        -> Dec (QuorumReached wallet signatures)


Then we guard the withdrawal operation with the instance of that predicate:

    withdrawFromWallet :
        (wallet : Var)
        -> (amount : NanoGram)
        -> (signatures : Vect (S k) PublicIdentity)
        -> (quorumReached : QuorumReached (MkMultiSigWalletState) signatures)
        -> {auto notExceedingTheBalanceProof : LTE amount balance}
        -> ST
            context
            ()
            [ wallet ::: State (MultiSigWalletState address balance owners quorum) :-> State (MultiSigWalletState address (balance-amount) owners quorum) ]

This effectively enforces, that `withdrawFromWallet` can only be used with prior check of `quorumReached`. Note, that since the state of the wallet is reflected in the type, we were able to specify, that withdrawal reduces the balance of the wallet.

See the "src/Playground.idr" for the sample flow. 


## Conclusion
In general, the approach of using dependent types language for smart contract developing looks very promising. However, I think, the current "state of the art" in this scientific domain is still not suitable for industry adoption. The amount of work even for the simple proofs is far too big and the tooling support is not satisfactory yet. 

Its not clear, why the simple TypeScript run-time check:
    
        const proof : boolean = this.owners.has(identity)

        if (proof) transfer.signedBy.add(identity)
    
should grow into the 2 pages of incomprehensible code. Of course, there are "erased" proofs like `Data.So`, but then the whole point of mathematical purity is lost.

So I think, the research in this area should be continued. And industry should guide the academia in the right direction, otherwise they will easily get lost in the type jungles. 

Big companies, like Telegram, are in the best position to lead such research activities. And it is very dissapointing to see that Telegram throws hundreds of $k at charting libraries (??) instead of, for example, sponsoring the (guided) Idris development. 

I would suggest, TON to support additional functional language with dependent types (Idris seems to be the only candidate) along with classic imperative ones.


About myself
------------

My open source projects: 

- [Chronograph], actively developed, being used by Bryntum in production

Chronograph can be characterized as:
- A system with the discrete time, in which a graph of atoms is advanced from previous state to a new one every "tick".
- An immutable graph, which provides such features as built-in undo/redo of the system state, cheap branching and cancelabale transactions
- In addition to synchronous calculations mode, Chronograph supports calculations defined as generator functions (which can be seen as delimited continuations). This is the most interesting property, in turn, it allows ChronoGraph to have the following novel capabilities for this class of libraries:
    -- Support for indefinitely deep data dependencies (the execution stack is externalized from the language into implementation and only limited by the memory). 
    -- Support for asynchronous calculations. 
    -- Possibility for running the calculations in breadth-first order, opposite to usual depth-first mode. This is important, if, for example, the underlaying storage layer is asynchronous (which is always desirable). In such case, we won't block the calculations by waiting for some value to arrive from storage layer, and continue other calculations in the meantime
- A graph of (mostly) pure calculation. A carefully chosen set of effects is allowed in atom calculations, which leads to more robust code.

- [Syncler], obsolete
- [Joose], obsolete

Github account:
- https://github.com/SamuraiJack

Blogs:
- https://medium.com/@nickolay8
- http://joose.it/blog/

Occasionally, I write posts for the blog of the company I'm working in:
- https://www.bryntum.com/blog/the-mixin-pattern-in-typescript-all-you-need-to-know
- https://www.bryntum.com/blog/writing-isomorphic-tests-with-siesta-5-2-0/


Telegram account:
- @matjek

Email:
- nickolay8@gmail.com


References
----------

[Chronograph]: https://github.com/bryntum/chronograph/tree/quark_master

[Syncler]: http://joose.it/blog/2011/04/08/syncler-distributed-applications-for-human-beings/

[Joose]: http://joose.it/

[TypeScript]: http://www.typescriptlang.org/

[Idris]: https://www.idris-lang.org/
