module MultiSigWallet

import Data.Vect
import Control.ST
import Util

%default total
%access public export

---------------------
NanoGram : Type
NanoGram = Nat

Address : Type
Address = Nat

PublicKey : Type
PublicKey = Nat

Guid : Type
Guid = Nat

Date : Type
Date = Nat

---------------------
record PublicIdentity where
    constructor MkPublicIdentity
    publicKey   : PublicKey


proofContra : (key1 : Nat) -> (key2 : Nat) -> (contra : key1 = key2 -> Void) -> (contra2 : MkPublicIdentity key1 = MkPublicIdentity key2) -> Void
proofContra key2 key2 contra Refl = contra Refl

DecEq PublicIdentity where
    decEq (MkPublicIdentity key1) (MkPublicIdentity key2) = case (decEq key1 key2) of
        (Yes prf) => case prf of
            Refl => Yes Refl
        (No contra) => No (proofContra key1 key2 contra)


---------------------
data MultiSigWalletState : (address : Address) -> (balance : NanoGram) -> (owners : Vect (S n) PublicIdentity) -> (quorum : NonZeroNat) -> Type where
    MkMultiSigWalletState : MultiSigWalletState address balance owners quorum

---------------------
SignatureBelongsToOwners : (owners : Vect (S n) PublicIdentity) -> (signature : PublicIdentity) -> Type
SignatureBelongsToOwners owners signature = Elem signature owners

---------------------
data QuorumReached :
    (wallet : MultiSigWalletState address balance owners quorum)
    -> (signatures : Vect (S k) PublicIdentity)
    -> Type
where
    IfAndOnlyIf :
        (wallet : MultiSigWalletState address balance owners quorum)
        -> (allSignaturesBelongsToOwners : ForEveryElementVect signatures (SignatureBelongsToOwners owners))
        -> (enoughSignatures : GTE (S k) (extractNonZeroNat quorum))
        -> QuorumReached wallet signatures

proofContrary :
    (wallet : MultiSigWalletState address balance owners quorum)
    -> (signatures : Vect (S k) PublicIdentity)
    -> (contra : ForEveryElementVect signatures (SignatureBelongsToOwners owners) -> Void)
    -> (quorumReached : QuorumReached wallet signatures) -> Void
proofContrary wallet signatures contra (IfAndOnlyIf wallet x {enoughSignatures}) = contra x


quorumReached :
    (wallet : MultiSigWalletState address balance owners quorum)
    -> (signatures : Vect (S k) PublicIdentity)
    -> {auto enoughSignatures : GTE (S k) (extractNonZeroNat quorum) }
    -> Dec (QuorumReached wallet signatures)
quorumReached wallet signatures {enoughSignatures} =
    case wallet of
        MkMultiSigWalletState =>
            let
                curriedIsElem = \el => isElem el owners
            in
                case (forEveryElementVect signatures (SignatureBelongsToOwners owners) curriedIsElem) of
                    (Yes prf) => Yes (IfAndOnlyIf MkMultiSigWalletState prf enoughSignatures)
                    (No contra) => No (proofContrary MkMultiSigWalletState signatures contra)


-------------------
createWallet :
    (address : Address)
    -> (quorum : NonZeroNat)
    -> (owners : Vect (S n) PublicIdentity)
    -> ST
        context
        Var
        [ add (State (MultiSigWalletState address 0 owners quorum)) ]

createWallet address quorum owners = new MkMultiSigWalletState

-------------------
depositToWallet :
    (wallet : Var)
    -> (amount : NanoGram)
    -> ST
        context
        ()
        [ wallet ::: State (MultiSigWalletState address 0 owners quorum) :-> State (MultiSigWalletState address amount owners quorum) ]

depositToWallet wallet amount = do
    walletState <- read wallet
    write wallet (MkMultiSigWalletState)

-------------------
getWalletBalance :
    (wallet : Var)
    -> ST
        context
        NanoGram
        [ wallet ::: State (MultiSigWalletState address balance owners quorum) ]

getWalletBalance wallet = do
    walletState <- read wallet
    pure balance

-------------------
withdrawFromWallet :
    (wallet : Var)
    -> (amount : NanoGram)
    -> (signatures : Vect (S k) PublicIdentity)
    -> (quorumReached : QuorumReached (MkMultiSigWalletState) signatures)
    -> {auto notExceedingTheBalanceProof : LTE amount balance}
    -> ST
        context
        ()
        [ wallet ::: State (MultiSigWalletState address balance owners quorum) :-> State (MultiSigWalletState address (balance-amount) owners quorum) ]

withdrawFromWallet wallet amount signatures quorumReached {notExceedingTheBalanceProof} = do
    walletState <- read wallet
    write wallet (MkMultiSigWalletState)
