import { Base } from "../class/Mixin.js";
//----------------------------------------------------------------------------------------------------------------------
export const PublicIdentity = (base) => class Identity extends base {
};
//----------------------------------------------------------------------------------------------------------------------
export const PrivateIdentity = (base) => class PrivateIdentity extends base {
};
//----------------------------------------------------------------------------------------------------------------------
export class Message extends Base {
}
export class MessageWalletTransfersState extends Message {
}
export class MessageTransferOrder extends Message {
}
export class MessageBoolean extends Message {
}
//----------------------------------------------------------------------------------------------------------------------
export class Account extends PrivateIdentity(PublicIdentity(Base)) {
    constructor() {
        super(...arguments);
        this.balance = 0;
    }
}
//----------------------------------------------------------------------------------------------------------------------
export class User extends Account {
}
//----------------------------------------------------------------------------------------------------------------------
export class SmartContract extends Account {
}
//----------------------------------------------------------------------------------------------------------------------
export class TransferOrder extends Base {
    constructor() {
        super(...arguments);
        this.signedBy = new Set();
    }
    sign(identity) {
        this.signedBy.add(identity);
    }
}
//----------------------------------------------------------------------------------------------------------------------
export class TransferOrderRequest extends Base {
    constructor() {
        super(...arguments);
        this.expiresAt = new Date(Date.now() + 3 * 60 * 60 * 1000);
    }
    isDataEqual(another) {
        return this.guid === another.guid
            && this.amount === another.amount
            && this.fromWallet === another.fromWallet
            && this.destination === another.destination
            && this.expiresAt.getTime() === another.expiresAt.getTime();
    }
}
//----------------------------------------------------------------------------------------------------------------------
export class MultiSigWallet extends SmartContract {
    constructor() {
        super(...arguments);
        this.stateClock = 0;
        this.transferOrders = new Map();
    }
    initialize(...args) {
        super.initialize(...args);
        if (!this.owners || !this.owners.size)
            throw new Error('Need at least 1 wallet owner');
        if (this.quorum == null)
            throw new Error('No quorum specified');
        if (this.quorum == 0)
            throw new Error('Quorum should bigger than 0');
        if (this.quorum > this.owners.size)
            throw new Error('Specified quorum is bigger than number of owners');
    }
    garbageCollect() {
        const now = new Date();
        this.transferOrders.forEach((order, key) => {
            if (order.request.expiresAt > now)
                this.transferOrders.delete(key);
        });
    }
    addOrMergeTransfer(transfer) {
        let existingTransfer = this.transferOrders.get(transfer.request.guid);
        if (existingTransfer) {
            if (!existingTransfer.request.isDataEqual(transfer.request))
                throw new Error('Transfer identity mismatch');
        }
        else {
            if (transfer.signedBy.size === 0)
                throw new Error('Need at least one signature for the transfer');
            existingTransfer = TransferOrder.new({ request: transfer.request });
            this.transferOrders.set(existingTransfer.request.guid, existingTransfer);
        }
        for (const identity of transfer.signedBy) {
            this.addSignature(existingTransfer, identity);
        }
        if (existingTransfer.signedBy.size >= this.quorum) {
            this.releaseTransfer(existingTransfer);
            this.transferOrders.delete(existingTransfer.request.guid);
            return MessageBoolean.new({ stateClock: this.stateClock, payload: true });
        }
        else {
            return MessageBoolean.new({ stateClock: this.stateClock, payload: false });
        }
    }
    addSignature(transfer, identity) {
        const proof = this.owners.has(identity);
        // the only way of extending the `signedBy` field of the `TransferOrder`
        // is guarded by the proof that identity is a wallet owner
        if (proof)
            transfer.signedBy.add(identity);
    }
    releaseTransfer(transfer) {
        if (transfer.request.amount > this.balance)
            throw new Error('Not enough balance to fulfill the transfer');
        this.balance -= transfer.request.amount;
    }
    getOrders() {
        return MessageWalletTransfersState.new({
            stateClock: this.stateClock,
            payload: Array.from(this.transferOrders.values())
        });
    }
    getOrdersSignedBy(identity) {
        return MessageWalletTransfersState.new({
            stateClock: this.stateClock,
            payload: Array.from(this.transferOrders.values()).filter(order => order.signedBy.has(identity))
        });
    }
    getOrdersNotSignedBy(identity) {
        return MessageWalletTransfersState.new({
            stateClock: this.stateClock,
            payload: Array.from(this.transferOrders.values()).filter(order => !order.signedBy.has(identity))
        });
    }
    getStateClock() {
        return Message.new({ stateClock: this.stateClock, payload: null });
    }
    receiveExternal(message) {
        if (message.stateClock !== this.stateClock)
            throw new Error('Message sent from the the stale state observation');
        this.garbageCollect();
        this.stateClock++;
        if (message.payload instanceof TransferOrder) {
            return this.addOrMergeTransfer(message.payload);
        }
        else {
            throw new Error("Unknown message type");
        }
    }
}
//----------------------------------------------------------------------------------------------------------------------
let counter = 0;
const newGuid = () => ++counter;
//----------------------------------------------------------------------------------------------------------------------
export const newTransferOrder = (me, fromWallet, amount, destination, expiresAt = new Date(Date.now() + 3 * 60 * 60 * 1000)) => {
    return TransferOrder.new({
        request: TransferOrderRequest.new({
            guid: newGuid(),
            amount,
            destination,
            fromWallet,
            expiresAt
        }),
        signedBy: new Set([me])
    });
};
