import { AnyConstructor, Base, Mixin } from "../class/Mixin.js"
import { Address, Guid, Id, NanoGram, Nat, PrivateKey, PublicKey, Signature } from "./Types.js"


//----------------------------------------------------------------------------------------------------------------------
export const PublicIdentity = <T extends AnyConstructor<object>>(base : T) =>

class Identity extends base {
    publicKey       : PublicKey
}

export type PublicIdentity = Mixin<typeof PublicIdentity>


//----------------------------------------------------------------------------------------------------------------------
export const PrivateIdentity = <T extends AnyConstructor<PublicIdentity>>(base : T) =>

class PrivateIdentity extends base {
    privateKey      : PrivateKey
}

export type PrivateIdentity = Mixin<typeof PrivateIdentity>



//----------------------------------------------------------------------------------------------------------------------
export class Message<Payload = any> extends Base {
    stateClock      : Nat

    payload         : Payload
}

export class MessageWalletTransfersState extends Message<TransferOrder[]> {}
export class MessageTransferOrder extends Message<TransferOrder> {}
export class MessageBoolean extends Message<boolean> {}


//----------------------------------------------------------------------------------------------------------------------
export class Account extends PrivateIdentity(PublicIdentity(Base)) {
    address         : Address

    balance         : NanoGram          = 0
}


//----------------------------------------------------------------------------------------------------------------------
export class User extends Account {
}


//----------------------------------------------------------------------------------------------------------------------
export class SmartContract extends Account {
}


//----------------------------------------------------------------------------------------------------------------------
export class TransferOrder extends Base {
    request         : TransferOrderRequest

    signedBy        : Set<PublicIdentity>     = new Set()


    sign (identity : PublicIdentity) {
        this.signedBy.add(identity)
    }
}


//----------------------------------------------------------------------------------------------------------------------
export class TransferOrderRequest extends Base {
    guid            : Guid

    fromWallet      : Account
    destination     : Account

    expiresAt       : Date              = new Date(Date.now() + 3 * 60 * 60 * 1000)

    amount          : NanoGram


    isDataEqual (another : TransferOrderRequest) : boolean {
        return this.guid === another.guid
            && this.amount === another.amount
            && this.fromWallet === another.fromWallet
            && this.destination === another.destination
            && this.expiresAt.getTime() === another.expiresAt.getTime()
    }
}


//----------------------------------------------------------------------------------------------------------------------
export class MultiSigWallet extends SmartContract {
    owners          : Set<PublicIdentity>

    quorum          : Nat

    stateClock      : Nat   = 0

    private transferOrders  : Map<Guid, TransferOrder> = new Map()


    initialize (...args) {
        super.initialize(...args)

        if (!this.owners || !this.owners.size) throw new Error('Need at least 1 wallet owner')

        if (this.quorum == null) throw new Error('No quorum specified')

        if (this.quorum == 0) throw new Error('Quorum should bigger than 0')

        if (this.quorum > this.owners.size) throw new Error('Specified quorum is bigger than number of owners')
    }


    garbageCollect () {
        const now   = new Date()

        this.transferOrders.forEach((order, key) => {
            if (order.request.expiresAt > now) this.transferOrders.delete(key)
        })
    }


    addOrMergeTransfer (transfer : TransferOrder) : MessageBoolean {
        let existingTransfer : TransferOrder     = this.transferOrders.get(transfer.request.guid)

        if (existingTransfer) {
            if (!existingTransfer.request.isDataEqual(transfer.request)) throw new Error('Transfer identity mismatch')
        } else {
            if (transfer.signedBy.size === 0) throw new Error('Need at least one signature for the transfer')

            existingTransfer = TransferOrder.new({ request : transfer.request })

            this.transferOrders.set(existingTransfer.request.guid, existingTransfer)
        }

        for (const identity of transfer.signedBy) {
            this.addSignature(existingTransfer, identity)
        }

        if (existingTransfer.signedBy.size >= this.quorum) {
            this.releaseTransfer(existingTransfer)

            this.transferOrders.delete(existingTransfer.request.guid)

            return MessageBoolean.new({ stateClock : this.stateClock, payload : true })
        } else {
            return MessageBoolean.new({ stateClock : this.stateClock, payload : false })
        }
    }


    private addSignature (transfer : TransferOrder, identity) {
        const proof : boolean = this.owners.has(identity)

        // the only way of extending the `signedBy` field of the `TransferOrder`
        // is guarded by the proof that identity is a wallet owner
        if (proof) transfer.signedBy.add(identity)
    }


    private releaseTransfer (transfer : TransferOrder) {
        if (transfer.request.amount > this.balance) throw new Error('Not enough balance to fulfill the transfer')

        this.balance -= transfer.request.amount
    }


    getOrders () : MessageWalletTransfersState {
        return MessageWalletTransfersState.new({
            stateClock      : this.stateClock,
            payload         : Array.from(this.transferOrders.values())
        })
    }


    getOrdersSignedBy (identity : PublicIdentity) : MessageWalletTransfersState {
        return MessageWalletTransfersState.new({
            stateClock      : this.stateClock,
            payload         : Array.from(this.transferOrders.values()).filter(order => order.signedBy.has(identity))
        })
    }


    getOrdersNotSignedBy (identity : PublicIdentity) : MessageWalletTransfersState {
        return MessageWalletTransfersState.new({
            stateClock      : this.stateClock,
            payload         : Array.from(this.transferOrders.values()).filter(order => !order.signedBy.has(identity))
        })
    }


    getStateClock () : Message {
        return Message.new({ stateClock : this.stateClock, payload : null })
    }


    receiveExternal (message : Message) : Message {
        if (message.stateClock !== this.stateClock) throw new Error('Message sent from the the stale state observation')

        this.garbageCollect()

        this.stateClock++

        if (message.payload instanceof TransferOrder) {
            return this.addOrMergeTransfer(message.payload)
        } else {
            throw new Error("Unknown message type")
        }
    }
}

//----------------------------------------------------------------------------------------------------------------------
let counter = 0

const newGuid = () => ++counter


//----------------------------------------------------------------------------------------------------------------------
export const newTransferOrder = (
    me : PublicIdentity, fromWallet : Account, amount : NanoGram, destination : Account,
    expiresAt : Date   = new Date(Date.now() + 3 * 60 * 60 * 1000)
) : TransferOrder => {
    return TransferOrder.new({
        request        : TransferOrderRequest.new({
            guid : newGuid(),
            amount,
            destination,
            fromWallet,
            expiresAt
        }),

        signedBy    : new Set([ me ])
    })
}
