import { MessageTransferOrder, MultiSigWallet, newTransferOrder, User } from "../../src/ton/MultiSigWallet.js"

declare const StartTest : any

StartTest(t => {

    let owner1 : User
    let owner2 : User
    let owner3 : User
    let destination : User

    let wallet : MultiSigWallet

    t.beforeEach(t => {
        owner1          = new User()
        owner2          = new User()
        owner3          = new User()
        destination     = new User()

        wallet          = MultiSigWallet.new({
            owners      : new Set([ owner1, owner2, owner3 ]),
            quorum      : 2,

            balance     : 100
        })
    })

    t.it('Should not be possible to instantiate malformed wallet', async t => {
        t.throwsOk(
            () => MultiSigWallet.new(),
            'Need at least 1 wallet owner'
        )

        t.throwsOk(
            () => MultiSigWallet.new({ owners : new Set() }),
            'Need at least 1 wallet owner'
        )

        t.throwsOk(
            () => MultiSigWallet.new({ owners : new Set([ owner1 ]) }),
            'No quorum specified'
        )

        t.throwsOk(
            () => MultiSigWallet.new({ owners : new Set([ owner1 ]), quorum : 0 }),
            'Quorum should bigger than 0'
        )

        t.throwsOk(
            () => MultiSigWallet.new({ owners : new Set([ owner1 ]), quorum : 2 }),
            'Specified quorum is bigger than number of owners'
        )
    })


    t.it('Should be possible to create and distribute transfer order to accumulate signatures', async t => {
        const transferOrder = newTransferOrder(owner1, wallet, 10, destination)

        t.isDeeply(transferOrder.signedBy, new Set([ owner1 ]), 'Signed by the transfer initiator')

        transferOrder.sign(owner2)

        t.isDeeply(transferOrder.signedBy, new Set([ owner1, owner2 ]), 'Signed by another owner')
    })


    t.it('Sending an order w/o quorum should just save it in the list of pending transfers', async t => {
        const transferOrder = newTransferOrder(owner1, wallet, 10, destination)

        const stateClock    = wallet.getStateClock().stateClock

        const message       = wallet.receiveExternal(MessageTransferOrder.new({
            stateClock      : stateClock,
            payload         : transferOrder
        }))

        t.is(message.payload, false, 'Transfer have not been released')

        const orders        = wallet.getOrders().payload

        t.is(orders.length, 1)
        t.ok(orders[ 0 ].request.isDataEqual(transferOrder.request), 'Order saved in the pending orders map')
    })


    t.it('Sending an order w/o quorum should just save it in the list of pending transfers', async t => {
        const transferOrder = newTransferOrder(owner1, wallet, 10, destination)

        const stateClock    = wallet.getStateClock().stateClock

        const message       = wallet.receiveExternal(MessageTransferOrder.new({
            stateClock      : stateClock,
            payload         : transferOrder
        }))

        t.is(message.payload, false, 'Transfer have not been released')
    })


    t.it('Sending an order w/ quorum should release it', async t => {
        const transferOrder = newTransferOrder(owner1, wallet, 10, destination)

        transferOrder.sign(owner2)

        const stateClock    = wallet.getStateClock().stateClock

        const initialBalance    = wallet.balance

        const message       = wallet.receiveExternal(MessageTransferOrder.new({
            stateClock      : stateClock,
            payload         : transferOrder
        }))

        t.is(message.payload, true, 'Transfer have been released')
        t.is(wallet.balance, initialBalance - transferOrder.request.amount, 'Funds have been transfered')
    })
})
